package subcluens

import data.Clustering._
import data.ConsensusSubspaceCluster
import subcluens.datastructure.DataTensor
import subcluens.operations.{FusionCols, FusionRows, FusionTubes, RegroupAll}

import scala.annotation.tailrec
import scala.concurrent.{Await, Future}
import scala.concurrent.duration.Duration

/**
  * This objects provides the facilities for the creation of meta clusters based on a decomposed data tensor
  */
object MetaClusterCreation {

  implicit val ec = scala.concurrent.ExecutionContext.Implicits.global


  def apply(caResult: DataTensor[Int, DimId, DpId]): Seq[ConsensusSubspaceCluster] = {
    caResult.dim1Indices.flatMap { clgIdx =>
      val group = caResult.caCells(clgIdx)
      //First we forcefully fuse the column and tube groups until only two groups remain (per dimension)
      val fourGroupsLeft = fusionDim2Dim3(caResult.copyWithNewCaCells(Vector(group)))
      //Next we try to fuse the result again, but know it also has to reduce the encoding costs
      val groupedTensor = totalFusionMinEncoding(fourGroupsLeft, List(fourGroupsLeft))
      //Find the cell with the hightest true rate
      val cell = groupedTensor.caCells.flatten.flatten.minBy(_.encodingCostsTrue)

      //Create a meta cluster based on the the data point and feature ids of the above found cell
      Seq(ConsensusSubspaceCluster(clgIdx, cell.dim1Ids.size, cell.dim2Ids, cell.dim3Ids))
    }.groupBy(x => (x.dataPoints, x.dimensions)).map(x => x._2.head).toSeq
  }

  /**
    * This function tries to fuse the groups of all dimensions, as long as it reduces the encoding costs.
    */
  @tailrec
  private def totalFusionMinEncoding(metaClusterT: DataTensor[Int, DimId, DpId], potentialClusters: List[DataTensor[Int, DimId, DpId]]): DataTensor[Int, DimId, DpId] = {
    if (metaClusterT.nrOf1DGroups > 2 || metaClusterT.nrOf2DGroups > 2 || metaClusterT.nrOf3DGroups > 2) {
      val rowFusion = Future {
        if (metaClusterT.nrOf1DGroups > 2) {
          Some(RegroupAll(FusionRows.performOperation(metaClusterT)))
        } else {
          None
        }
      }
      val colFusion = Future {
        if (metaClusterT.nrOf2DGroups > 2) {
          Some(RegroupAll(FusionCols.performOperation(metaClusterT)))
        } else {
          None
        }
      }
      val tubeFusion = Future {
        if (metaClusterT.nrOf3DGroups > 2) {
          Some(RegroupAll(FusionTubes.performOperation(metaClusterT)))
        } else {
          None
        }
      }
      val f = Future.sequence {
        Seq(rowFusion, colFusion, tubeFusion)
      }.map(_.flatten.minBy(_.encodingCosts))
      val winner = Await.result(f, Duration.Inf)
      totalFusionMinEncoding(winner, winner :: potentialClusters)
    } else {
      potentialClusters.minBy(_.encodingCosts)
    }
  }

  /**
    * This function fuses the groups of cols and tubes forcefully as long as the number of groups is greater than two.
    */
  @tailrec
  final def fusionDim2Dim3(metaClusterT: DataTensor[Int, DimId, DpId]): DataTensor[Int, DimId, DpId] = {
    val GROUPS_PER_DIM = 2

    if (metaClusterT.nrOf2DGroups > GROUPS_PER_DIM || metaClusterT.nrOf3DGroups > GROUPS_PER_DIM) {
      val colFusion = Future {
        if (metaClusterT.nrOf2DGroups > GROUPS_PER_DIM)
          Some(RegroupAll(FusionCols.performOperation(metaClusterT)))
        else
          None
      }
      val tubesFusion = Future {
        if (metaClusterT.nrOf3DGroups > GROUPS_PER_DIM)
          Some(RegroupAll(FusionTubes.performOperation(metaClusterT)))
        else
          None
      }
      val f = Future.sequence {
        Seq(tubesFusion, colFusion)
      }.map(_.flatten.minBy(_.encodingCosts))
      val winner = Await.result(f, Duration.Inf)
      fusionDim2Dim3(winner)
    } else {
      RegroupAll(metaClusterT)
    }
  }
}
