package subcluens

import com.typesafe.scalalogging.LazyLogging
import subcluens.datastructure.DataTensor
import subcluens.operations._

import scala.annotation.tailrec
import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration.Duration
import scala.concurrent.{Await, Future}

/**
  * This object contains the function which performs the modified version of the three dimensional cross association, as described in the paper.
  * For each iteration it performs all six possible operations (split and merge for each dimension) in combination with a re
  */
object ModifiedCrossAssociation3d extends LazyLogging {

  @tailrec
  private def runner(dm: DataTensor[Int, Int, Int], counter: Int, optionalMaxRowGroups: Option[Int]): DataTensor[Int, Int, Int] = {
    logger.info("round: " + counter + " current costs: " + dm.encodingCosts)

    //The quick and dirty way to parallelism
    val _rowSplit = Future {
      if (optionalMaxRowGroups.forall(_ > dm.nrOf1DGroups)) {
        val start = System.currentTimeMillis()
        val splitted = SplitRow.performOperation(dm)
        val res = if (splitted ne dm) RegroupAll(splitted) else dm
        val end = System.currentTimeMillis()
        logger.info(s"row split took ${end - start}: ${res.encodingCosts}")
        res
      } else {
        logger.info(s"rows: reached maximal number")
        dm
      }
    }
    val _colSplit = Future {
      val start = System.currentTimeMillis()
      val splitted = SplitCol.performOperation(dm)
      val res = if (splitted ne dm) RegroupAll(splitted) else dm
      val end = System.currentTimeMillis()
      logger.info(s"col split took ${end - start}: ${res.encodingCosts}")
      res
    }
    val _tubeSplit = Future {
      val start = System.currentTimeMillis()
      val splitted = SplitTube.performOperation(dm)
      val res = if (splitted ne dm) RegroupAll(splitted) else dm
      val end = System.currentTimeMillis()
      logger.info(s"tube split took ${end - start}: ${res.encodingCosts}")
      res
    }
    val _rowFusion = Future {
      val start = System.currentTimeMillis()
      val splitted = FusionRows.performOperation(dm)
      val res = if (splitted ne dm) RegroupAll(splitted) else dm
      val end = System.currentTimeMillis()
      logger.info(s"row fusion took ${end - start}: ${res.encodingCosts}")
      res
    }
    val _colFusion = Future {
      val start = System.currentTimeMillis()
      val splitted = FusionCols.performOperation(dm)
      val res = if (splitted ne dm) RegroupAll(splitted) else dm
      val end = System.currentTimeMillis()
      logger.info(s"col fusion took ${end - start}: ${res.encodingCosts}")
      res
    }
    val _tubeFusion = Future {
      val start = System.currentTimeMillis()
      val splitted = FusionTubes.performOperation(dm)
      val res = if (splitted ne dm) RegroupAll(splitted) else dm
      val end = System.currentTimeMillis()
      logger.info(s"tube fusion took ${end - start}: ${res.encodingCosts}")
      res
    }

    val f = Future.sequence {
      Seq(_rowSplit, _colSplit, _tubeSplit, _rowFusion, _colFusion, _tubeFusion)
    }.map { l => l.minBy(_.encodingCosts) }
    val winner = Await.result(f, Duration.Inf)

    if (dm.encodingCosts > winner.encodingCosts) {
      runner(winner, counter + 1, optionalMaxRowGroups)
    } else {
      dm
    }
  }

  def run(dm: DataTensor[Int, Int, Int], optionalMaxRowGroups: Option[Int] = None): DataTensor[Int, Int, Int] = runner(dm, 0, optionalMaxRowGroups)
}
