package subcluens

import com.typesafe.scalalogging.LazyLogging
import data.Clustering._
import subcluens.datastructure.DataTensor

/**
  * This object represents the full algorithm SubCluEns as described in the paper.
  * It decomposes the given data tensor until a local optimum is reached.
  * Then it creates the meta-clusters as described in the paper.
  */
object SubCluEns extends LazyLogging {
  /**
    *
    * @param initialDt A initial data tensor with one group per dimension
    * @param permutationSeed A integer value which is used as a seeding value for the permutation of all ids (for rows,cols, tubes)
    * @param optionalMaxClusterGroups An optional parameter, if set to a positive integer value, the number of cluster groups will not exceed this value
    * @return The created meta-clusters
    */
  def apply(initialDt: DataTensor[Int, DimId, DpId], permutationSeed: Int, optionalMaxClusterGroups: Option[Int] = None) = {

    //We create a random permutation of all ids for each dimension. This ensures that we run into different local minima for each run
    val (tensor, caMapping) = GenerateDataTensor.shuffleIds(initialDt, permutationSeed)

    //We decompose the tensor using the modified cross association as described in the paper
    val decomposedTensor = ModifiedCrossAssociation3d.run(tensor, optionalMaxClusterGroups)
    logger.info(s"found local optimum with ${decomposedTensor.nrOf1DGroups} rows, ${decomposedTensor.nrOf2DGroups} cols, ${decomposedTensor.nrOf3DGroups} tubes,")

    //We revert the permutation of the of the ids for each dimension and sort the groups by size
    val deshuffledTensor = GenerateDataTensor.shuffleBackIds(decomposedTensor, caMapping)

    //We create the meta clusters as described in the paper
    val metaClusters = MetaClusterCreation.apply(deshuffledTensor)

    //We return the meta clusters as well as the encoding costs of the decomposed tensor
    (metaClusters, decomposedTensor.encodingCosts)
  }
}
