package subcluens.datastructure

/**
  * This trait represents an abstract "view"  onto the tensor.
  * It represents a slice for one group
  * For each dimension we provide some utility functions which enable the reuse for different decomposition operations
  */
trait TensorSegment[A, X, Y, Z] {
  def dimension: Int

  def groupNr: Int

  def groupIds: Set[A]

  def dimAIndices: Range

  def dimBIndices: Range

  def caCell(a: Int, b: Int): CaCell[X, Y, Z]


  protected def trueCountForDimIdx(ca: CaCell[X, Y, Z], idx: A): Int

  def getTrueCountForEntry(a: Int, b: Int)(idx: A): Int = trueCountForDimIdx(caCell(a, b), idx)

  def trueEncodingCosts(a: Int, b: Int): Double = caCell(a, b).encodingCostsTrue


  /**
    * Determines the sum of term 6 for all cells for each group
    */
  lazy val entropyPerSlice = {
    var sum = 0d
    for {
      a <- dimAIndices
      b <- dimBIndices
    } {
      sum += caCell(a, b).dataEncodingCosts
    }
    sum / groupIds.size
  }

  /**
    * Determines the encoding costs of this segment (nrofTrues- and data encoding costs, as described in paper for the fusion operation)
    *
    */
  lazy val segmentEncodingCosts: Double = {
    var costs = 0d
    for {
      a <- dimAIndices
      b <- dimBIndices
    } {
      val ca = caCell(a, b)
      costs += ca.dataEncodingCosts + ca.encodingCostNrOfTrues
    }
    costs
  }

  /**
    * Determines the encoding costs of this segment in combination with the other segment
    * (nrofTrues- and data encoding costs, as described in paper for the fusion operation)
    */
  def combinedSegmentEncodingCosts(otherSeg: TensorSegment[A, X, Y, Z]): Double = {
    var costs = 0d
    for {
      a <- dimAIndices
      b <- dimBIndices
    } {
      val ca1 = caCell(a, b)
      val ca2 = otherSeg.caCell(a, b)
      val trueCount = (ca1.trueCount + ca2.trueCount).toDouble
      val size = (ca1.size + ca2.size).toDouble
      val (encodingCostsTrue,encodingCostsFalse) = EncodingCosts3d.determinePerEventEncodingCosts(trueCount, size)
      val encodingCostNrOfTrues = EncodingCosts3d.caCellNrOfTruesCosts(size)
      val dataEncodingCosts = EncodingCosts3d.determineEncodingCosts(trueCount, size - trueCount, encodingCostsTrue, encodingCostsFalse)
      costs += dataEncodingCosts + encodingCostNrOfTrues
    }
    costs
  }
}

/**
  * Represents a slice of a row group
  */
case class TensorDim1Segment[X, Y, Z](groupNr: Int, caCells: IndexedSeq[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]]) extends TensorSegment[X, X, Y, Z] {
  override val dimension = 1
  @transient
  lazy val groupIds: Set[X] = caCells(groupNr)(0)(0).dim1Ids

  def dimAIndices: Range = caCells(0).indices

  def dimBIndices: Range = caCells(0)(0).indices

  def caCell(a: Int, b: Int): CaCell[X, Y, Z] = caCells(groupNr)(a)(b)

  protected def trueCountForDimIdx(ca: CaCell[X, Y, Z], idx: X) = ca.trueCountForDim1Idx(idx)
}


/**
* Represents a slice of a column group
*/
case class TensorDim2Segment[X, Y, Z](groupNr: Int, caCells: IndexedSeq[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]]) extends TensorSegment[Y, X, Y, Z] {
  override val dimension = 2
  @transient
  lazy val groupIds: Set[Y] = caCells(0)(groupNr)(0).dim2Ids

  def dimAIndices: Range = caCells.indices

  def dimBIndices: Range = caCells(0)(0).indices

  def caCell(a: Int, b: Int): CaCell[X, Y, Z] = caCells(a)(groupNr)(b)

  protected def trueCountForDimIdx(ca: CaCell[X, Y, Z], idx: Y) = ca.trueCountForDim2Idx(idx)
}

/**
  * Represents a slice of a tube group
  */
case class TensorDim3Segment[X, Y, Z](groupNr: Int, caCells: IndexedSeq[IndexedSeq[IndexedSeq[CaCell[X, Y, Z]]]]) extends TensorSegment[Z, X, Y, Z] {
  override val dimension = 3
  @transient
  lazy val groupIds: Set[Z] = caCells(0)(0)(groupNr).dim3Ids

  def dimAIndices: Range = caCells.indices

  def dimBIndices: Range = caCells(0).indices

  def caCell(a: Int, b: Int): CaCell[X, Y, Z] = caCells(a)(b)(groupNr)

  protected def trueCountForDimIdx(ca: CaCell[X, Y, Z], idx: Z) = ca.trueCountForDim3Idx(idx)
}